var ExtractionReg = function()
{
	var exreg = {};
	var debug = 0;
	
	// QIAcube positions to be used for a given number of samples
	// Ordered from 1 to 12 samples
	var QIACUBE_POS = [
		[1],
		[1, 7],
		[1, 5, 9],
		[1, 2, 7, 8],
		[1, 2, 5, 8, 9],
		[1, 2, 3, 7, 8, 9],
		[1, 2, 3, 6, 7, 9, 10],
		[1, 2, 3, 4, 7, 8, 9, 10],
		[1, 2, 3, 5, 6, 7, 9, 10, 11],
		[1, 2, 3, 4, 5, 7, 8, 9, 10, 11],
		[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11],
		[1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
	];
	
	var selectedLysates;
	var allLysatesAreReExtraction;
	
	var latestRna;
	var latestLysate;
	
	exreg.REAGENTS = [
		'rltPlusBuffer', 'betaMercaptoetanol', 'reagentDx', 'qiaShredderColumn',
		'allPrepMiniKit', 'etoh70', 'bufferRw1', 'bufferRpe', 'bufferAw1', 
		'bufferAw2', 'bufferEb', 'rnaseFreeWater', 'allPrepDnaColumn', 'rneasyColumn' 
		];

	
	var DEFAULT_LYSATE_VOLUME = 700;
	var DEFAULT_LYSATE_USED = 350;
	var DEFAULT_DNA_VOLUME = 95;
	var DEFAULT_RNA_VOLUME = 53;
	var DEFAULT_FT_VOLUME = 550;
	var MINIMAL_PRE_NORMALIZE_QUANTITY = 0; // µg, =0 since we no longer handle NotEnoughQuantity in this wizard

	var lysatesIsValid = false;
	var qiacubeDateIsAfterLysisDate = true;
	var qiacubeRunNoIsValid = false;
	var qiacubePositionsAreValid = false;
	var nanoDropFileIsValid = false;
	var createdRna = [];
	
	// Page initialization
	exreg.initPage = function()
	{
		// Step 1
		Events.addEventHandler('step-1', 'wizard-validate', exreg.validateStep1);
		Events.addEventHandler('lysates', 'change', exreg.lysatesOnChange);
		Events.addEventHandler('reextraction', 'change', exreg.lysatesOnChange);
		Buttons.addClickHandler('downloadLabels', exreg.downloadLabels);
		
		// Step 2
		Events.addEventHandler('step-2', 'wizard-initialize', exreg.initializeStep2);
		Events.addEventHandler('step-2', 'wizard-validate', exreg.validateStep2);
		
		Events.addEventHandler('lysisDate', 'change', Wizard.validateDate);
		Events.addEventHandler('qiacubeDate', 'change', Wizard.validateDate);
		Events.addEventHandler('lysisDate', 'change', exreg.qiacubeDateOnChange);
		Events.addEventHandler('qiacubeDate', 'change', exreg.qiacubeDateOnChange);
		Events.addEventHandler('qiacubeRunNo', 'change', exreg.qiacubeRunNoOnChange);
		
		Buttons.addClickHandler('btnUseLatestLysateReagents', exreg.useLatestReagents);
		Buttons.addClickHandler('btnUseLatestRnaReagents', exreg.useLatestReagents);
		
		// Step 3
		Events.addEventHandler('step-3', 'wizard-initialize', exreg.initializeStep3);
		Events.addEventHandler('step-3', 'wizard-validate', exreg.validateStep3);
	
		Buttons.addClickHandler('downloadSampleIdFile', exreg.downloadSampleIdFile);
		Buttons.addClickHandler('btnFakeNanoFropFile', exreg.generateFakeNanoDropFile);
		Events.addEventHandler('nanoDropFile', 'change', exreg.nanoDropFileOnChange);
		
		// Navigation
		Buttons.addClickHandler('gocancel', Wizard.cancelWizard);
		Buttons.addClickHandler('gorestart', Wizard.restartWizard);
		Buttons.addClickHandler('gonext', Wizard.goNextOnClick);
		Buttons.addClickHandler('goregister', Wizard.goRegister);
		Buttons.addClickHandler('rnaQcWizard', exreg.goRnaQc);
		
		// Final registration
		Events.addEventHandler('wizard', 'wizard-submit', exreg.submit);

		var url = '../Extraction.servlet?ID='+App.getSessionId();
		url += '&cmd=GetUnprocessedLysates&includeReprocess=1';		
		Wizard.showLoadingAnimation('Loading Lysate extracts...');
		Wizard.asyncJsonRequest(url, exreg.initializeStep1);
	}
	
	exreg.initializeStep1 = function(response)
	{
		var frm = document.forms['reggie'];
		var lysates = response.lysates;
		
		var lysateList = frm.lysates;
		var reextractList = frm.reextraction;
		if (lysates.length > 0)
		{
			// Check if we have any lysate reserved by the current user
			// and count the number of YellowLabel specimen
			var currentUser = Data.get(lysateList, 'current-user');
			var hasReserved = false;
			var numYellow = 0;
			for (var i=0; i < lysates.length; i++)
			{
				var lysate = lysates[i];
				if (lysate.ReservedBy == currentUser) hasReserved = true;
				if (lysate.specimen.YellowLabel != null) numYellow++;
				lysate.isReProcess = lysate.AutoProcessing == 'ReProcess';
			}

			var numSelected = 0;
			var maxNonYellow = 12-numYellow;
			for (var i=0; i < lysates.length; i++)
			{
				var lysate = lysates[i];
				var isYellow = lysate.specimen.YellowLabel != null;
				var name = (i+1) + ': ' + Strings.encodeTags(lysate.name);
				if (lysate.bioWell)
				{
					name += ' -- ' + Strings.encodeTags(lysate.bioWell.bioPlate.name + ' (' + lysate.bioWell.location+')');
				}
				if (lysate.ReservedBy)
				{
					name += ' ['+lysate.ReservedBy+']';
				}
				
				// We only auto-select items reserved by the current user (if it has any)
				// and never any items reserved by other users
				var selected = false;
				if (numSelected < 12)
				{
					if (hasReserved)
					{
						selected = lysate.ReservedBy == currentUser;
					}
					else
					{
						selected = isYellow || maxNonYellow > 0;
					}
					if (selected) 
					{
						numSelected++;
						if (!isYellow) maxNonYellow--;
					}
				}
				
				var option = new Option(name, lysate.id, selected, selected);
				if (isYellow) option.className = 'yellow';
				option.lysate = lysate;
				if (lysate.isReProcess)
				{
					reextractList.options[reextractList.length] = option;
				}
				else
				{
					lysateList.options[lysateList.length] = option;
				}
			}
			if (lysateList.length == 0) Doc.hide('regularFlowRow');
			if (reextractList.length > 0) Doc.show('reExtractionRow');
			exreg.lysatesOnChange();
		}
		else
		{
			Wizard.setFatalError('No Lysate available for processing.');
			return;
		}

		Doc.show('step-1');
		Doc.show('gonext');
	}

	
	exreg.lysatesOnChange = function()
	{
		lysatesIsValid = false;
		Wizard.setInputStatus('lysates');
		Wizard.setInputStatus('reextraction');
		
		var frm = document.forms['reggie'];
		
		var numLysates = 0;
		var numReextraction = 0;
		for (var i = 0; i < frm.lysates.length; i++)
		{
			if (frm.lysates[i].selected) numLysates++;
		}
		for (var i = 0; i < frm.reextraction.length; i++)
		{
			if (frm.reextraction[i].selected) numReextraction++;
		}

		if (numReextraction > 0)
		{
			Doc.removeClass('downloadLabels', 'disabled');
		}
		else
		{
			Doc.addClass('downloadLabels', 'disabled');
		}
		
		if (numReextraction == 0 && numLysates == 0)
		{
			Wizard.setInputStatus('lysates', 'invalid', 'At least one lysate must be selected');
			Wizard.setInputStatus('reextraction', 'invalid', 'At least one lysate must be selected');
			return;
		}

		if (numReextraction+numLysates > 12)
		{
			if (numLysates > 0) Wizard.setInputStatus('lysates', 'invalid', 'Must not select more than 12 items.');
			if (numReextraction > 0) Wizard.setInputStatus('reextraction', 'invalid', 'Must not select more than 12 items.');
			return;
		}
		
		lysatesIsValid = true;
		Wizard.setInputStatus('lysates', 'valid');
		Wizard.setInputStatus('reextraction', 'valid');
	}

	exreg.downloadLabels = function()
	{
		var frm = document.forms['reggie'];
		
		var lysates = [];
		for (var i = 0; i < frm.reextraction.length; i++)
		{
			var lysate = frm.reextraction[i].lysate;
			if (frm.reextraction[i].selected)
			{
				lysates[lysates.length] = lysate.id;
			}
		}
		
		var frm = document.forms['reggie'];
		var url = '../Extraction.servlet?ID='+App.getSessionId();
		url += '&cmd=DownloadLabelFile';
		url += '&lysates=' + encodeURIComponent(lysates.join(','));
		window.open(url);
		
	}
	
	exreg.validateStep1 = function(event)
	{
		if (!lysatesIsValid)
		{
			event.preventDefault();
		}
	}
	
	exreg.initializeStep2 = function()
	{
		var frm = document.forms['reggie'];
		frm.lysates.disabled = true;
		
		selectedLysates = [];
		allLysatesAreReExtraction = true;
		var lysateIds = [];
		for (var i = 0; i < frm.lysates.length; i++)
		{
			if (frm.lysates[i].selected) 
			{
				var lys = frm.lysates[i].lysate;
				selectedLysates[selectedLysates.length] = lys;
				lysateIds[lysateIds.length] = lys.id;
				allLysatesAreReExtraction = false;
			}
		}
		for (var i = 0; i < frm.reextraction.length; i++)
		{
			if (frm.reextraction[i].selected) 
			{
				var lys = frm.reextraction[i].lysate;
				selectedLysates[selectedLysates.length] = lys;
				lysateIds[lysateIds.length] = lys.id;
			}
		}
		
		var url = '../Extraction.servlet?ID='+App.getSessionId();
		url += '&cmd=GetChildItemNames';
		url += '&lysates='+lysateIds.join(',');
		Wizard.showLoadingAnimation('Loading child item names...');
		Wizard.asyncJsonRequest(url, exreg.childNamesLoaded);
		
	}
		
	exreg.childNamesLoaded = function(response)
	{
		var frm = document.forms['reggie'];
		var childNames = response.childNames;
		
		// Create child RNA/DNA/FlowThrough items
		for (var i = 0; i < selectedLysates.length; i++)
		{
			var lysate = selectedLysates[i];
				
			var rna = {};
			var dna = {};
			var ft = {};
			// DNA
			dna.name = childNames[i].dna;
			dna.volume = DEFAULT_DNA_VOLUME;
				
			// RNA
			rna.name = childNames[i].rna;
			rna.volume = DEFAULT_RNA_VOLUME;
				
			// FlowThrough
			ft.name = childNames[i].ft;
			ft.volume = DEFAULT_FT_VOLUME;
			
			lysate.rna = rna;
			lysate.dna = dna;
			lysate.ft = ft;
		}
		
		Wizard.setCurrentStep(2);
		Doc.show('gocancel');
		Doc.show('gonext');
		
		if (!allLysatesAreReExtraction)
		{
			Reggie.loadProtocols('SAMPLE_HANDLING_PROTOCOL', 'lysisProtocol');
		}
		Doc.addClass('dnaProtocol', 'list-loading');
		frm.dnaProtocol[0] = new Option('loading...', '');
		Doc.addClass('rnaProtocol', 'list-loading');
		frm.rnaProtocol[0] = new Option('loading...', '');
		Doc.addClass('ftProtocol', 'list-loading');
		frm.ftProtocol[0] = new Option('loading...', '');
		var url = '../Protocol.servlet?ID='+App.getSessionId();
		url += '&cmd=GetProtocols&subtype=EXTRACTION_PROTOCOL';
		Wizard.asyncJsonRequest(url, exreg.protocolsLoaded);

		if (allLysatesAreReExtraction)
		{
			Doc.element('only-re-extraction').innerHTML = '(Disabled since all selected lysates are for re-extraction)';
			frm.lysisDate.value = '';
			frm.lysisDate.disabled = true;
			frm.lysisProtocol.disabled = true;
			frm.rltPlusBuffer.disabled = true;
			frm.betaMercaptoetanol.disabled = true;
			frm.reagentDx.disabled = true;
			frm.qiaShredderColumn.disabled = true;
			frm.qiacubeDate.focus();
		}
		else
		{
			frm.lysisDate.focus();
			Events.sendChangeEvent('lysisDate');
		}
		Events.sendChangeEvent('qiacubeDate');
		
		var url = '../Extraction.servlet?ID='+App.getSessionId();
		url += '&cmd=GetLatestReagentLotNumbers';		
		Wizard.showLoadingAnimation('Loading recent reagent lot numbers...');
		Wizard.asyncJsonRequest(url, exreg.reagentsLoaded);
	}
	
	exreg.protocolsLoaded = function(response)
	{
		var frm = document.forms['reggie'];
		var protocols = response.protocols;

		Doc.removeClass('dnaProtocol', 'list-loading');
		Doc.removeClass('rnaProtocol', 'list-loading');
		Doc.removeClass('ftProtocol', 'list-loading');
		frm.dnaProtocol.length = 0;
		frm.rnaProtocol.length = 0;
		frm.ftProtocol.length = 0;
		for (var i = 0; i < protocols.length; i++)
		{
			var name = protocols[i].name;
			var id = protocols[i].id;
			frm.dnaProtocol[frm.dnaProtocol.length] = new Option(name, id, false, name == 'AllprepCustomPartA');
			frm.rnaProtocol[frm.rnaProtocol.length] = new Option(name, id, false, name == 'AllprepMod1partC');
			frm.ftProtocol[frm.ftProtocol.length] = new Option(name, id, false, name == 'AllprepMod1partB');
		}
		
		frm.dnaProtocol[frm.dnaProtocol.length] = new Option('- none -', '');
		frm.rnaProtocol[frm.rnaProtocol.length] = new Option('- none -', '');
		frm.ftProtocol[frm.ftProtocol.length] = new Option('- none -', '');
	}
	
	exreg.reagentsLoaded = function(response)
	{
		latestRna = response.rna;
		latestLysate = response.lysate;
		
		if (!latestRna) return;
		
		var haveLatestLysateReagents = false;
		var haveLatestRnaReagents = false;
		for (var rgNo = 0; rgNo < exreg.REAGENTS.length; rgNo++)
		{
			var rg = exreg.REAGENTS[rgNo];
			if (latestLysate[rg])
			{
				haveLatestLysateReagents = true;
			}
			if (latestRna[rg])
			{
				haveLatestRnaReagents = true;
			}
		}
		
		if (!allLysatesAreReExtraction && haveLatestLysateReagents)
		{
			Doc.show('useLatestLysateReagents');
			Doc.element('latestLysisDate').innerHTML = '(Lysis date ' + Reggie.reformatDate(latestLysate.lysisDate) + ')';
		}
		else
		{
			latestLysate = null;
		}
		
		if (haveLatestRnaReagents)
		{
			Doc.show('useLatestRnaReagents');
			Doc.element('latestQiacubeDate').innerHTML = '(QiaCube run ' + Reggie.reformatDate(latestRna.QiaCubeDate) + ')';
		}
		else
		{
			latestRna = null;
		}
	}
	
	exreg.useLatestReagents = function(event)
	{
		var which = Data.get(event.currentTarget, 'which');
		
		var latestItem = which == 'rna' ? latestRna : latestLysate;
		if (!latestItem) return;
		
		var frm = document.forms['reggie'];
		for (var rgNo = 0; rgNo < exreg.REAGENTS.length; rgNo++)
		{
			var rg = exreg.REAGENTS[rgNo];
			if (latestItem[rg])
			{
				frm[rg].value = latestItem[rg];
			}
		}
	}
	
	exreg.qiacubeDateOnChange = function()
	{
		var frm = document.forms['reggie'];
		
		var lysisDateIsValid = Wizard.isValid('lysisDate');
		var qiacubeDateIsValid = Wizard.isValid('qiacubeDate');
		
		if (!lysisDateIsValid || !qiacubeDateIsValid) return;

		var lysisDate = Dates.parseString(frm.lysisDate.value, 'yyyyMMdd');
		var qiacubeDate = Dates.parseString(frm.qiacubeDate.value, 'yyyyMMdd');
		if (lysisDate > qiacubeDate)
		{
			Wizard.setInputStatus('qiacubeDate', 'invalid', 'Qiacube date is before Lysis date');
			qiacubeDateIsAfterLysisDate = false;
			return;
		}

		qiacubeDateIsAfterLysisDate = true;
		
		Wizard.setInputStatus('qiacubeDate', 'valid');
		var url = '../Extraction.servlet?ID='+App.getSessionId();
		url += '&cmd=GetNextQiacubeRunNo&qiacubeDate='+encodeURIComponent(frm.qiacubeDate.value);
		Wizard.showLoadingAnimation('Loading Qiacube run number...');
		Wizard.asyncJsonRequest(url, exreg.gotNextRunNumber);
	}

	exreg.gotNextRunNumber = function(response)
	{
		var frm = document.forms['reggie'];
		frm.qiacubeRunNo.value = response.nextQiacubeRunNo;
		Events.sendChangeEvent('qiacubeRunNo');
	}

	exreg.qiacubeRunNoOnChange = function()
	{
		var frm = document.forms['reggie'];
		var runNumber = frm.qiacubeRunNo.value;
		qiacubeRunNoIsValid = false;

		if (runNumber == '')
		{
			Wizard.setInputStatus('qiacubeRunNo', 'invalid', 'Missing');
			return;
		}
		if (!(parseInt(runNumber, 10) >= 1))
		{
			Wizard.setInputStatus('qiacubeRunNo', 'invalid', 'Must be at least 1');
			return;		
		}
		
		Wizard.setInputStatus('qiacubeRunNo', 'valid');
		qiacubeRunNoIsValid = true;
	}

	
	exreg.validateStep2 = function(event)
	{
		var valid = true;
		valid &= Wizard.isValid('lysisDate');
		valid &= Wizard.isValid('qiacubeDate');
		valid &= qiacubeRunNoIsValid;
		
		if (!valid) event.preventDefault();
	}

	
	exreg.initializeStep3 = function()
	{
		var frm = document.forms['reggie'];
		
		var html = '<table id="detailsTable" class="step-form">';
		// First header row
		html += '<thead>';
		html += '<tr>';
		html += '<th>Specimen</th>';
		html += '<th class="dottedleft">QIAcube</th>';
		html += '<th colspan="3" class="dottedleft">Lysate</th>';
		html += '<th colspan="3" class="dottedleft">DNA</th>';
		html += '<th colspan="5" class="dottedleft">RNA</th>';
		html += '<th class="dottedleft">FT</th>';
		html += '<th class="dottedleft comment">Comment</th>';
		html += '<th></th>';
		html += '</tr>';
		// Second header row
		html += '<tr>';
		html += '<th></th>';
		html += '<th class="dottedleft">position</th>';
		html += '<td class="dottedleft">Storage</td><td>Total</td><td>Used</td>';
		html += '<td class="dottedleft">NanoDrop</td><td>Vol.</td><td>Conc.</td>';
		html += '<td class="dottedleft">NanoDrop</td><td>Vol.</td><td>Conc.</td><td>Quantity</td><td><span id="toggle-normalize-1" class="link" title="Toggle all">Pre-</span></td>';
		html += '<td class="dottedleft">Vol.</td>';
		html += '<td class="dottedleft comment"></td>';
		html += '<td></td>';
		html += '</tr>';
		// Third header row
		html += '<tr>';
		html += '<th></th>';
		html += '<th class="dottedleft"></th>';
		html += '<td class="dottedleft">box</td><td>(µl)</td><td>(µl)</td>';
		html += '<td class="dottedleft">pos</td><td>(µl)</td><td>(ng/µl)</td>';
		html += '<td class="dottedleft">pos</td><td>(µl)</td><td>(ng/µl)</td><td>(µg)</td><td><span id="toggle-normalize-2" class="link" title="Toggle all">normalize</span></td>';
		html += '<td class="dottedleft">(µl)</td>';
		html += '<td class="dottedleft comment"></td>';
		html += '<td></td>';
		html += '</tr>';
		html += '</thead>';
		html += '<tbody>';
		
		// Set default values for DNA/RNA extraction
		var QPOS = QIACUBE_POS[selectedLysates.length-1];
		for (var i = 0; i < selectedLysates.length; i++)
		{
			var lysate = selectedLysates[i];
			var specimen = lysate.specimen;
			var isYellow = specimen.YellowLabel != null;
			var id = lysate.id;
			if (!lysate.isReProcess)
			{
				lysate.originalQuantity = DEFAULT_LYSATE_VOLUME;
			}
			lysate.processedQuantity = DEFAULT_LYSATE_USED;
			lysate.qiacubePosition = QPOS[i];
			lysate.originalQiacubePosition = QPOS[i];
				
			var img = isYellow ? '<img src="../images/yellow-label.png">' : '';
			html += '<tr class="highlight'+(isYellow ? ' yellow-specimen' : '')+'">';
			html += '<th class="specimen if-yellow">'+img+Strings.encodeTags(specimen.name) + '</th>';
			html += '<td id="qiacubePosition'+id+'" class="dottedleft"></td>';
			html += '<td class="dottedleft">'+Strings.encodeTags(lysate.bioWell.bioPlate.name + '[' + lysate.bioWell.location)+']</td>';
			html += '<td id="lysateTotal'+id+'"></td>';
			html += '<td id="lysateUsed'+id+'"></td>';
			html += '<td id="dnaPos'+id+'" class="dottedleft"></td>';
			html += '<td id="dnaVol'+id+'"></td>';
			html += '<td id="dnaConc'+id+'" class="bg-filled-50 italic"></td>';
			html += '<td id="rnaPos'+id+'" class="dottedleft"></td>';
			html += '<td id="rnaVol'+id+'"></td>';
			html += '<td id="rnaConc'+id+'" class="bg-filled-50 italic"></td>';
			html += '<td id="rnaQuantity'+id+'" class="bg-filled-50 italic"></td>';
			html += '<td><input type="checkbox" name="rnaPreNormalize'+id+'" id="rnaPreNormalize'+id+'" disabled>';
			html += '<span id="rnaPreNormalize'+id+'.status" class="status">&nbsp;</span></td>';
			html += '<td id="ftVol'+id+'" class="dottedleft"></td>';
			html += '<td id="comments'+id+'" class="dottedleft comment"></td>';
			html += '<td><div class="button basicbutton interactable edit-details" style="width: 5em;" data-lysate-id="'+id+'">Edit&hellip;</div></td>';
			html += '</tr>';
		}
		html += '</tbody>';
		html += '</table>';
		Doc.element('detailsSection').innerHTML = html;
		
		var buttons = Doc.element('detailsSection').getElementsByClassName('edit-details');
		for (var i = 0; i < buttons.length; i++)
		{
			Buttons.addClickHandler(buttons[i], exreg.editDnaRnaDetails);
		}
		
		Events.addEventHandler('toggle-normalize-1', 'click', exreg.toggleNormalize);
		Events.addEventHandler('toggle-normalize-2', 'click', exreg.toggleNormalize);
		
		exreg.updateDetailsTable();
		
		Wizard.setCurrentStep(3);
		Doc.show('gocancel');
		Doc.show('goregister');
		
		Wizard.setInputStatus('nanoDropFile', 'invalid', 'No file selected');
		Wizard.keepSessionAlive();
		
		if (debug && location.protocol != 'https:') 
		{
			Wizard.showDebugTools();
		}
	}
	
	exreg.toggleNormalize = function()
	{
		var frm = document.forms['reggie'];
		var check = null;
		for (var i = 0; i < selectedLysates.length; i++)
		{
			var id = selectedLysates[i].id;
			if (!frm['rnaPreNormalize'+id].disabled)
			{
				if (check == null) check = !frm['rnaPreNormalize'+id].checked;
				frm['rnaPreNormalize'+id].checked = check;
			}
		}
	}
	
	exreg.calculateNanoDropPositions = function()
	{
		var tmp = [];
		for (var i = 0; i < selectedLysates.length; i++)
		{
			var o = {};
			o.index = i;
			o.qiapos = selectedLysates[i].qiacubePosition;
			tmp[tmp.length] = o;
		}
		tmp.sort(exreg.sortByQiaCubePosition);
		
		var ndCol = 1;
		var ndRow = 0;
		var rnaOffset = selectedLysates.length > 7 ? 2 : 1;

		for (var i = 0; i < tmp.length; i++)
		{
			var lysate = selectedLysates[tmp[i].index];
			lysate.nanoDropOrder = i;
			lysate.dna.well = Reggie.wellToAlpha(ndRow) + ndCol;
			lysate.rna.well = Reggie.wellToAlpha(ndRow) + (ndCol+rnaOffset);
			ndRow++;
			if (ndRow == 8)
			{
				ndRow = 0;
				ndCol++;
			}
		}
	}
	
	exreg.sortByQiaCubePosition = function(a, b)
	{
		return a.qiapos == b.qiapos ? a.index - b.index : a.qiapos - b.qiapos;
	}
	
	exreg.updateDetailsTable = function()
	{
		var frm = document.forms['reggie'];
		var hasQiacubePosition = [];
		var duplicatePosition = 0;
		
		exreg.calculateNanoDropPositions();
		
		for (var i = 0; i < selectedLysates.length; i++)
		{
			var lysate = selectedLysates[i];
			var id = lysate.id;
			Doc.element('qiacubePosition'+id).innerHTML = lysate.qiacubePosition;
			Doc.addOrRemoveClass('qiacubePosition'+id, 'nondefault', lysate.qiacubePosition != lysate.originalQiacubePosition);
			
			if (hasQiacubePosition[lysate.qiacubePosition])
			{
				duplicatePosition = lysate.qiacubePosition;
			}
			else
			{
				hasQiacubePosition[lysate.qiacubePosition] = true;
			}
			
			Doc.element('lysateTotal'+id).innerHTML = Numbers.formatNumber(lysate.originalQuantity || lysate.remainingQuantity, 0);
			Doc.addOrRemoveClass('lysateTotal'+id, 'nondefault', !lysate.isReProcess && lysate.originalQuantity != DEFAULT_LYSATE_VOLUME);

			Doc.element('lysateUsed'+id).innerHTML = lysate.processedQuantity;
			Doc.addOrRemoveClass('lysateUsed'+id, 'nondefault', lysate.processedQuantity != DEFAULT_LYSATE_USED);
		
			Doc.element('dnaPos'+id).innerHTML = lysate.dna.well;
			Doc.addOrRemoveClass('dnaPos'+id, 'nondefault', lysate.qiacubePosition != lysate.originalQiacubePosition);
			Doc.element('dnaVol'+id).innerHTML = lysate.dna.volume;
			Doc.addOrRemoveClass('dnaVol'+id, 'nondefault', lysate.dna.volume != DEFAULT_DNA_VOLUME);
			var dnaConc = Doc.element('dnaConc'+id);
			dnaConc.innerHTML = Reggie.formatNumber(lysate.dna.ndConc) || '-';
			Doc.addOrRemoveClass(dnaConc, 'low-conc', lysate.dna.ndConc < 1);
			dnaConc.title = '260/280=' + lysate.dna.nd260by280 + '\n260/230=' + lysate.dna.nd260by230;
		
			Doc.element('rnaPos'+id).innerHTML =  lysate.rna.well;
			Doc.addOrRemoveClass('rnaPos'+id, 'nondefault', lysate.qiacubePosition != lysate.originalQiacubePosition);
			Doc.element('rnaVol'+id).innerHTML = lysate.rna.volume;
			Doc.addOrRemoveClass('rnaVol'+id, 'nondefault', lysate.rna.volume != DEFAULT_RNA_VOLUME);
			var rnaConc = Doc.element('rnaConc'+id);
			rnaConc.innerHTML = Reggie.formatNumber(lysate.rna.ndConc) || '-';
			Doc.addOrRemoveClass(rnaConc, 'low-conc', lysate.rna.ndConc < 1);
			rnaConc.title = '260/280=' + lysate.rna.nd260by280 + '\n260/230=' + lysate.rna.nd260by230;
			
			lysate.rna.quantity = Math.max(0, lysate.rna.ndConc * lysate.rna.volume / 1000);
			var enoughQuantity = lysate.rna.quantity > MINIMAL_PRE_NORMALIZE_QUANTITY;
			lysate.rna.flag = enoughQuantity ? null : 'NotEnoughRemainingQuantity';
			
			frm['rnaPreNormalize'+id].disabled = !enoughQuantity;
			Doc.element('rnaQuantity'+id).innerHTML = Reggie.formatNumber(lysate.rna.quantity || null, null, 2) || '-';
			frm['rnaPreNormalize'+id].checked = enoughQuantity;
			var warning = lysate.rna.quantity && !enoughQuantity;
			Wizard.setInputStatus('rnaPreNormalize'+id, warning ? 'flag' : '', warning ? 'Not enough quantity' : '');
			
			Doc.element('ftVol'+id).innerHTML = lysate.ft.volume;
			Doc.addOrRemoveClass('ftVol'+id, 'nondefault', lysate.ft.volume != DEFAULT_FT_VOLUME);
		
			var cmnt = Doc.element('comments'+id);
			var tooltip = [];
			var comment = [];
			if (lysate.comment || lysate.dna.comment || lysate.rna.comment || lysate.ft.comment)
			{
				comment[0] = 'Yes';
				if (lysate.comment) tooltip[tooltip.length] = 'Lysate: ' + lysate.comment;
				if (lysate.dna.comment) tooltip[tooltip.length] = 'DNA: ' + lysate.dna.comment;
				if (lysate.rna.comment) tooltip[tooltip.length] = 'RNA: ' + lysate.rna.comment;
				if (lysate.ft.comment) tooltip[tooltip.length] = 'FlowThrough: ' + lysate.ft.comment;
			}
			
			var modifiedReagents = false;
			for (var rgNo = 0; rgNo < exreg.REAGENTS.length; rgNo++)
			{
				var rg = exreg.REAGENTS[rgNo];
				if (lysate[rg]) 
				{
					tooltip[tooltip.length] = Doc.element(rg+'.title').innerHTML + ': ' + Strings.encodeTags(lysate[rg]);
					modifiedReagents = true;
				}
			}
			if (modifiedReagents)
			{
				comment[comment.length] = 'Reagents';
			}
			
			if (tooltip.length > 0)
			{
				cmnt.innerHTML = comment.join(', ');
				Doc.addClass(cmnt, 'nondefault');
				cmnt.title = tooltip.join('\n');
			}
			else
			{
				cmnt.innerHTML = '';
				Doc.removeClass(cmnt, 'nondefault');
				cmnt.title = null;
			}
		}
		
		if (duplicatePosition)
		{
			Wizard.setInputStatus('downloadSampleIdFile', 'invalid', 'Duplicate Qiacube position: ' + duplicatePosition);
			qiacubePositionsAreValid = false;
		}
		else
		{
			Wizard.setInputStatus('downloadSampleIdFile');
			qiacubePositionsAreValid = true;
		}
	}

	exreg.editDnaRnaDetails = function(event)
	{
		var lysateId = Data.int(event.currentTarget, 'lysate-id');
		var index = -1;
		for (var i = 0; i < selectedLysates.length; i++)
		{
			if (selectedLysates[i].id == lysateId)
			{
				index = i;
				break;
			}
		}
		
		var frm = document.forms['reggie'];
		var url = 'edit_dnarna_details.jsp?ID='+App.getSessionId();
		url += '&index='+index;
		url += '&lysate=' + encodeURIComponent(selectedLysates[index].name);
		Dialogs.openPopup(url, 'EditDnaRnaDetails'+index, 750, 500);	
	}
	
	exreg.getSelectedLysate = function(index)
	{
		return selectedLysates[index];
	}
	
	exreg.downloadSampleIdFile = function()
	{
		var frm = document.forms['reggie'];

		if (!qiacubePositionsAreValid)
		{
			Forms.showNotification('downloadSampleIdFile', 'QIAcube positions are not valid. Please fix before downloading the file.');
			return;
		}
		
		var rnaNames = [];
		var dnaNames = [];
		var frm = document.forms['reggie'];
		var url = '../Extraction.servlet?ID='+App.getSessionId();
		url += '&cmd=DownloadNanoDropSampleIdFile';
		url += '&qiacubeDate='+encodeURIComponent(frm.qiacubeDate.value);
		url += '&qiacubeRunNo='+encodeURIComponent(frm.qiacubeRunNo.value);
		
		var selected = [];
		for (var i = 0; i < selectedLysates.length; i++)
		{
			var lysate = selectedLysates[i];
			dnaNames[lysate.nanoDropOrder] = lysate.dna.name;
			rnaNames[lysate.nanoDropOrder] = lysate.rna.name;
		}
		url += '&rnaNames=' + encodeURIComponent(rnaNames.join(','));
		url += '&dnaNames=' + encodeURIComponent(dnaNames.join(','));
		window.open(url);
	}

	exreg.nanoDropFileOnChange = function()
	{
		var frm = document.forms['reggie'];
		var file = frm.nanoDropFile.files[0];
		
		var reader = new FileReader();
		reader.onload = exreg.nanoDropFileLoaded;
		reader.readAsText(file);
	}

	exreg.nanoDropFileLoaded = function(event)
	{
		exreg.parseNanoDropFile(event.target.result);
	}


	exreg.parseNanoDropFile = function(data)
	{
		var frm = document.forms['reggie'];
		nanoDropFileIsValid = false;

		try
		{
			// Reset the existing values
			for (var i = 0; i < selectedLysates.length; i++)
			{
				var lysate = selectedLysates[i];
				lysate.dna.ndConc = null;
				lysate.dna.nd260by280 = null;
				lysate.dna.nd260by230 = null;
				lysate.rna.ndConc = null;
				lysate.rna.nd260by280 = null;
				lysate.rna.nd260by230 = null;
			}
			
			var lines = data.split(/[\n\r]+/);
			var columns = lines[0].split(/ *\t */);
			
			var sampleIndex = exreg.findColumn(columns, 'Sample ID');
			if (sampleIndex == -1) return;
			
			var wellIndex = exreg.findColumn(columns, 'Well');
			if (wellIndex == -1) return;
			
			var ndConcIndex = exreg.findColumn(columns, 'Conc.');
			if (ndConcIndex == -1) return;
			
			var nd260by280Index = exreg.findColumn(columns, '260/280');
			if (nd260by280Index == -1) return;
			
			var nd260by230Index = exreg.findColumn(columns, '260/230');
			if (nd260by230Index == -1) return;
		
			var numImported = 0;		
			for (var i = 1; i < lines.length; i++)
			{
				if (lines[i].match(/^\s*$/)) continue; // Ignore empty lines
				
				var row = lines[i].split(/ *\t */);
				if (row.length != columns.length) 
				{
					Wizard.setInputStatus('nanoDropFile', 'invalid', 'On line ' + (i+1) + '; Unexpected number of columns: ' + row.length);
					return;
				}
				
				var sampleId = row[sampleIndex];
				if (!sampleId || sampleId == 'blank') continue; // with the next line
		
				var rnaOrDna = exreg.findRnaOrDna(sampleId, i+1);
				if (!rnaOrDna) return;  // This is an error condition
				
				if (rnaOrDna.well != row[wellIndex])
				{
					Wizard.setInputStatus('nanoDropFile', 'invalid', 'On line ' + (i+1) + '; Expected to find "' + sampleId + '" in well [' + rnaOrDna.well + '] not ['  + row[wellIndex] + ']');
					return;
				}
				
				var ndConc = exreg.getNumber(row[ndConcIndex]);
				if (isNaN(ndConc))
				{
					Wizard.setInputStatus('nanoDropFile', 'invalid', 'On line ' + (i+1) + '; No valid NDConc value for "' + sampleId + '": ' + row[ndConcIndex]);
					return;
				}
				
				// If there are multiple entries for the same item, keep the largest NDConc
				var isFirstValue = rnaOrDna.ndConc == null;
				if (isFirstValue || ndConc > rnaOrDna.ndConc)
				{
					var nd260by280 = exreg.getNumber(row[nd260by280Index]);
					if (isNaN(nd260by280))
					{
						Wizard.setInputStatus('nanoDropFile', 'invalid', 'On line ' + (i+1) + '; No valid 260/280 value for "' + sampleId + '": ' + row[nd260by280Index]);
						return;
					}
					
					var nd260by230 = exreg.getNumber(row[nd260by230Index]);
					if (isNaN(nd260by230))
					{
						Wizard.setInputStatus('nanoDropFile', 'invalid', 'On line ' + (i+1) + '; No valid 260/230 value for "' + sampleId + '": '+row[nd260by230Index]);
						return;
					}
					
					// Store values
					rnaOrDna.ndConc = ndConc;
					rnaOrDna.nd260by280 = nd260by280;
					rnaOrDna.nd260by230 = nd260by230;

					// Only update counter for the first value (not when replacing a lower value)
					if (isFirstValue)
					{
						numImported++;
					}
				}
				
			}
		
			if (numImported != selectedLysates.length*2)
			{
				Wizard.setInputStatus('nanoDropFile', 'invalid', 'Could not find values for all '+(selectedLysates.length*2)+' DNA/RNA samples: ' + numImported);
				return;
			}

			Wizard.setInputStatus('nanoDropFile', 'valid');
			nanoDropFileIsValid = true;
		}
		finally
		{
			exreg.updateDetailsTable();
		}
		
	}

	exreg.findColumn = function(columns, col)
	{
		var colIndex = columns.indexOf(col);
		if (colIndex == -1)
		{
			Wizard.setInputStatus('nanoDropFile', 'invalid', 'On line 1; Can\'t find "' + col + '" column');
		}
		return colIndex;
	}
	
	exreg.getNumber = function(sValue)
	{
		return parseFloat(sValue.replace(/,/, '.'));
	}

	exreg.findRnaOrDna = function(sampleId, lineNo)
	{
		var frm = document.forms['reggie'];
		for (var i = 0; i < selectedLysates.length; i++)
		{
			var lysate = selectedLysates[i];
			if (lysate.rna.name == sampleId)
			{
				return lysate.rna;
			}
			else if (lysate.dna.name == sampleId)
			{
				return lysate.dna;
			}
		}
		
		Wizard.setInputStatus('nanoDropFile', 'invalid', 'On line ' + lineNo + '; Can\'t find RNA/DNA with name "' + sampleId + '"');
		return null;
	}
	
	
	exreg.generateFakeNanoDropFile = function()
	{
		if (location.protocol == 'https:')
		{
			Forms.showNotification('btnFakeNanoFropFile', 'For security reasons, this function can\'t be used on HTTPS servers.');
			return;
		}
		
		var FILE_HEADER = ['Well', 'Sample ID', 'Conc.', '260/280', '260/230'];
		var data = FILE_HEADER.join('\t')+'\n';
		
		var frm = document.forms['reggie'];
		for (var i = 0; i < selectedLysates.length; i++)
		{
			var lysate = selectedLysates[i];
			var line = [];
			// DNA 
			line[0] = lysate.dna.well;
			line[1] = lysate.dna.name;
			line[2] = Numbers.formatNumber(Math.random() > 0.9 ? -Math.random() : Math.random()*200+5, 2);
			line[3] = Numbers.formatNumber(Math.random()*.5+2, 2);
			line[4] = Numbers.formatNumber(Math.random()+1, 2);
			data += line.join('\t') + '\n';
			// RNA
			line[0] = lysate.rna.well;
			line[1] = lysate.rna.name;
			line[2] = Numbers.formatNumber(Math.random() > 0.9 ? -Math.random() : Math.random()*200+5, 2);
			line[3] = Numbers.formatNumber(Math.random()*.5+2, 2);
			line[4] = Numbers.formatNumber(Math.random()+1, 2);
			data += line.join('\t') + '\n';
		}
		exreg.parseNanoDropFile(data);
	}

	exreg.validateStep3 = function(event)
	{
		var valid = true;
		valid &= nanoDropFileIsValid;
		valid &= qiacubePositionsAreValid;
		
		if (!valid) event.preventDefault();
	}


	exreg.submit = function()
	{
		var frm = document.forms['reggie'];
		
		var submitInfo = {};
		submitInfo.lysates = [];
		
		for (var i = 0; i < selectedLysates.length; i++)
		{
			var lysate = selectedLysates[i];
			lysate.rna.preNormalize = frm['rnaPreNormalize'+lysate.id].checked;
			submitInfo.lysates[submitInfo.lysates.length] = lysate;
		}
		
		// Lysis
		submitInfo.lysisDate = frm.lysisDate.value;
		submitInfo.lysisProtocol = parseInt(frm.lysisProtocol.value, 10);
		
		// Qiacube
		submitInfo.qiacubeDate = frm.qiacubeDate.value;
		submitInfo.qiacubeRunNo = parseInt(frm.qiacubeRunNo.value, 10);
		submitInfo.qiacubeOperator = frm.qiacubeOperator.value;
		submitInfo.dnaProtocol = parseInt(frm.dnaProtocol.value, 10);
		submitInfo.rnaProtocol = parseInt(frm.rnaProtocol.value, 10);
		submitInfo.ftProtocol = parseInt(frm.ftProtocol.value, 10);

		// Reagents
		for (var i = 0; i < exreg.REAGENTS.length; i++)
		{
			var rg = exreg.REAGENTS[i];
			submitInfo[rg] = frm[rg].value;
		}
		
		var url = '../Extraction.servlet?ID='+App.getSessionId();
		url += '&cmd=RegisterDnaRnaFt';
		Wizard.showLoadingAnimation('Performing registration...');
		Wizard.asyncJsonRequest(url, exreg.submissionResults, 'POST', JSON.stringify(submitInfo));
	}
	
	exreg.submissionResults = function(response)
	{
		Wizard.showFinalMessage(response.messages);
		Doc.show('gorestart');
		if (response.createdRna && response.createdRna.length > 0)
		{
			createdRna = response.createdRna;
			Doc.show('rnaQcWizard');	
		}
	}

	exreg.goRnaQc = function()
	{
		var url = 'rnaqc_aliquot.jsp?ID='+App.getSessionId();
		for (var i = 0; i < createdRna.length; i++)
		{
			url += '&rna='+createdRna[i];
		}
		location.href = url;
	}

	
	return exreg;
}();

Doc.onLoad(ExtractionReg.initPage);

